"""
Directed Acyclic Graphs (DAGs) may be used to construct complex workflows for
scikit-learn estimators. As the name suggests, data may only flow in one
direction and can't go back on itself to a previously run step.
"""
from collections import UserDict
from copy import deepcopy
from inspect import signature
from itertools import chain
from typing import Iterable

import networkx as nx
import numpy as np
from joblib import Parallel, delayed
from scipy.sparse import dok_matrix, issparse
from skdag.dag._render import DAGRenderer
from skdag.dag._utils import (
    _format_output,
    _in_notebook,
    _is_pandas,
    _is_passthrough,
    _is_predictor,
    _is_transformer,
    _stack,
)
from sklearn.base import clone
from sklearn.exceptions import NotFittedError
from sklearn.utils import Bunch, _print_elapsed_time, _safe_indexing, deprecated
from sklearn.utils._tags import _safe_tags
from sklearn.utils.metaestimators import _BaseComposition, available_if
from sklearn.utils.validation import check_is_fitted, check_memory

__all__ = ["DAG", "DAGStep"]


def _get_columns(X, dep, cols, is_root, axis=1):
    if callable(cols):
        # sklearn.compose.make_column_selector
        cols = cols(X)

    if not is_root:
        # The DAG will prepend output columns with the step name, so add this in to any
        # dep columns if missing. This helps keep user-provided deps readable.
        if isinstance(cols, str):
            cols = cols if cols.startswith(f"{dep}__") else f"{dep}__{cols}"
        elif isinstance(cols, Iterable):
            orig = cols
            cols = []
            for col in orig:
                if isinstance(col, str):
                    cols.append(col if col.startswith(f"{dep}__") else f"{dep}__{col}")
                else:
                    cols.append(col)

    return _safe_indexing(X, cols, axis=axis)


def _stack_inputs(dag, X, node):
    # For root nodes, the dependency is just the node name itself.
    deps = {node.name: None} if node.is_root else node.deps

    cols = [
        _get_columns(X[dep], dep, cols, node.is_root, axis=1)
        for dep, cols in deps.items()
    ]

    to_stack = [
        # If we sliced a single column from an input, reshape it to a 2d array.
        col.reshape(-1, 1)
        if col is not None and deps[dep] is not None and col.ndim < 2
        else col
        for col, dep in zip(cols, deps)
    ]

    X_stacked = _stack(to_stack, axis=node.axis)

    return X_stacked


def _leaf_estimators_have(attr, how="all"):
    """Check that leaves have `attr`.
    Used together with `avaliable_if` in `DAG`."""

    def check_leaves(self):
        # raises `AttributeError` with all details if `attr` does not exist
        failed = []
        for leaf in self.leaves_:
            try:
                _is_passthrough(leaf.estimator) or getattr(leaf.estimator, attr)
            except AttributeError:
                failed.append(leaf.estimator)

        if (how == "all" and failed) or (
            how == "any" and len(failed) != len(self.leaves_)
        ):
            raise AttributeError(
                f"{', '.join([repr(type(est)) for est in failed])} "
                f"object(s) has no attribute '{attr}'"
            )
        return True

    return check_leaves


def _transform_one(transformer, X, weight, allow_predictor=True, **fit_params):
    if _is_passthrough(transformer):
        res = X
    elif allow_predictor and not hasattr(transformer, "transform"):
        for fn in ["predict_proba", "decision_function", "predict"]:
            if hasattr(transformer, fn):
                res = getattr(transformer, fn)(X)
                if res.ndim < 2:
                    res = res.reshape(-1, 1)
                break
        else:
            raise AttributeError(
                f"'{type(transformer).__name__}' object has no attribute 'transform'"
            )
    else:
        res = transformer.transform(X)
    # if we have a weight for this transformer, multiply output
    if weight is not None:
        res = res * weight

    return res


def _fit_transform_one(
    transformer,
    X,
    y,
    weight,
    message_clsname="",
    message=None,
    allow_predictor=True,
    **fit_params,
):
    """
    Fits ``transformer`` to ``X`` and ``y``. The transformed result is returned
    with the fitted transformer. If ``weight`` is not ``None``, the result will
    be multiplied by ``weight``.
    """
    with _print_elapsed_time(message_clsname, message):
        failed = False
        if _is_passthrough(transformer):
            res = X
        elif hasattr(transformer, "fit_transform"):
            res = transformer.fit_transform(X, y, **fit_params)
        elif hasattr(transformer, "transform"):
            res = transformer.fit(X, y, **fit_params).transform(X)
        elif allow_predictor:
            for fn in ["predict_proba", "decision_function", "predict"]:
                if hasattr(transformer, fn):
                    res = getattr(transformer.fit(X, y, **fit_params), fn)(X)
                    if res.ndim < 2:
                        res = res.reshape(-1, 1)
                    break
            else:
                failed = True
                res = None

            if res is not None and res.ndim < 2:
                res = res.reshape(-1, 1)
        else:
            failed = True

        if failed:
            raise AttributeError(
                f"'{type(transformer).__name__}' object has no attribute 'transform'"
            )

    if weight is not None:
        res = res * weight

    return res, transformer


def _parallel_fit(dag, step, Xin, Xs, y, fit_transform_fn, memory, **fit_params):
    transformer = step.estimator

    if step.deps:
        X = _stack_inputs(dag, Xs, step)
    else:
        # For root nodes, the destination rather than the source is
        # specified.
        # X = Xin[step.name]
        X = _stack_inputs(dag, Xin, step)

    clsname = type(dag).__name__
    with _print_elapsed_time(clsname, dag._log_message(step)):
        if transformer is None or transformer == "passthrough":
            Xt, fitted_transformer = X, transformer
        else:
            if hasattr(memory, "location") and memory.location is None:
                # we do not clone when caching is disabled to
                # preserve backward compatibility
                cloned_transformer = transformer
            else:
                cloned_transformer = clone(transformer)

            # Fit or load from cache the current transformer
            Xt, fitted_transformer = fit_transform_fn(
                cloned_transformer,
                X,
                y,
                None,
                message_clsname=clsname,
                message=dag._log_message(step),
                **fit_params,
            )

    Xt = _format_output(Xt, X, step)

    return Xt, fitted_transformer


def _parallel_transform(dag, step, Xin, Xs, transform_fn, **fn_params):
    transformer = step.estimator
    if step.deps:
        X = _stack_inputs(dag, Xs, step)
    else:
        # For root nodes, the destination rather than the source is
        # specified.
        X = _stack_inputs(dag, Xin, step)
        # X = Xin[step.name]

    clsname = type(dag).__name__
    with _print_elapsed_time(clsname, dag._log_message(step)):
        if transformer is None or transformer == "passthrough":
            Xt = X
        else:
            # Fit or load from cache the current transformer
            Xt = transform_fn(
                transformer,
                X,
                None,
                message_clsname=clsname,
                message=dag._log_message(step),
                **fn_params,
            )

    Xt = _format_output(Xt, X, step)

    return Xt


def _parallel_fit_leaf(dag, leaf, Xts, y, **fit_params):
    with _print_elapsed_time(type(dag).__name__, dag._log_message(leaf)):
        if leaf.estimator == "passthrough":
            fitted_estimator = leaf.estimator
        else:
            Xt = _stack_inputs(dag, Xts, leaf)
            fitted_estimator = leaf.estimator.fit(Xt, y, **fit_params)

    return fitted_estimator


def _parallel_execute(
    dag, leaf, fn, Xts, y=None, fit_first=False, fit_params=None, fn_params=None
):
    with _print_elapsed_time("DAG", dag._log_message(leaf)):
        Xt = _stack_inputs(dag, Xts, leaf)
        fit_params = fit_params or {}
        fn_params = fn_params or {}
        if leaf.estimator == "passthrough":
            Xout = Xt
        elif fit_first and hasattr(leaf.estimator, f"fit_{fn}"):
            Xout = getattr(leaf.estimator, f"fit_{fn}")(Xt, y, **fit_params)
        else:
            if fit_first:
                leaf.estimator.fit(Xt, y, **fit_params)

            est_fn = getattr(leaf.estimator, fn)
            if "y" in signature(est_fn).parameters:
                Xout = est_fn(Xt, y=y, **fn_params)
            else:
                Xout = est_fn(Xt, **fn_params)

        Xout = _format_output(Xout, Xt, leaf)

    fitted_estimator = leaf.estimator

    return Xout, fitted_estimator


class DAGStep:
    """
    A single estimator step in a DAG.

    Parameters
    ----------
    name : str
        The reference name for this step.
    estimator : estimator-like
        The estimator (transformer or predictor) that will be executed by this step.
    deps : dict
        A map of dependency names to columns. If columns is ``None``, then all input
        columns will be selected.
    dataframe_columns : list of str or "infer" (optional)
        Either a hard-coded list of column names to apply to any output data, or the
        string "infer", which means the column outputs will be assumed to match the
        column inputs if the output is 2d and not already a dataframe, the estimator is
        a transformer, and the final axis dimensions match the inputs. Otherwise the
        column names will be assumed to be the step name + index if the output is not
        already a dataframe. If set to ``None`` or inference is not possible, the
        outputs will be left unmodified.
    axis : int, default = 1
        The strategy for merging inputs if there is more than upstream dependency.
        ``axis=0`` will assume all inputs have the same features and stack the rows
        together; ``axis=1`` will assume each input provides different features for the
        same samples.
    """

    def __init__(self, name, estimator, deps, dataframe_columns, axis=1):
        self.name = name
        self.estimator = estimator
        self.deps = deps
        self.dataframe_columns = dataframe_columns
        self.axis = axis
        self.index = None
        self.is_root = False
        self.is_leaf = False
        self.is_fitted = False

    def __repr__(self):
        return f"{type(self).__name__}({repr(self.name)}, {repr(self.estimator)})"


class DAG(_BaseComposition):
    """
    A Directed Acyclic Graph (DAG) of estimators, that itself implements the estimator
    interface.

    A DAG may consist of a simple chain of estimators (being exactly equivalent to a
    :mod:`sklearn.pipeline.Pipeline`) or a more complex path of dependencies. But as the
    name suggests, it may not contain any cyclic dependencies and data may only flow
    from one or more start points (roots) to one or more endpoints (leaves).

    Parameters
    ----------

    graph : :class:`networkx.DiGraph`
        A directed graph with string node IDs indicating the step name. Each node must
        have a ``step`` attribute, which contains a :class:`skdag.dag.DAGStep`.

    memory : str or object with the joblib.Memory interface, default=None
        Used to cache the fitted transformers of the DAG. By default, no caching is
        performed. If a string is given, it is the path to the caching directory.
        Enabling caching triggers a clone of the transformers before fitting. Therefore,
        the transformer instance given to the DAG cannot be inspected directly. Use the
        attribute ``named_steps`` or ``steps`` to inspect estimators within the
        pipeline. Caching the transformers is advantageous when fitting is time
        consuming.

    n_jobs : int, default=None
        Number of jobs to run in parallel. ``None`` means 1 unless in a
        :obj:`joblib.parallel_backend` context.

    verbose : bool, default=False
        If True, the time elapsed while fitting each step will be printed as it is
        completed.

    Attributes
    ----------

    graph_ : :class:`networkx.DiGraph`
        A read-only view of the workflow.

    classes_ : ndarray of shape (n_classes,)
        The classes labels. Only exists if the last step of the pipeline is a
        classifier.

    n_features_in_ : int
        Number of features seen during :term:`fit`. Only defined if all of the
        underlying root estimators in `graph_` expose such an attribute when fit.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Only defined if the underlying
        estimators expose such an attribute when fit.

    See Also
    --------
    :class:`skdag.DAGBuilder` : Convenience utility for simplified DAG construction.

    Examples
    --------

    The simplest DAGs are just a chain of singular dependencies. These DAGs may be
    created from the :meth:`skdag.dag.DAG.from_pipeline` method in the same way as a
    DAG:

    >>> from sklearn.decomposition import PCA
    >>> from sklearn.impute import SimpleImputer
    >>> from sklearn.linear_model import LogisticRegression
    >>> dag = DAG.from_pipeline(
    ...     steps=[
    ...         ("impute", SimpleImputer()),
    ...         ("pca", PCA()),
    ...         ("lr", LogisticRegression())
    ...     ]
    ... )
    >>> print(dag.draw().strip())
    o    impute
    |
    o    pca
    |
    o    lr

    For more complex DAGs, it is recommended to use a :class:`skdag.dag.DAGBuilder`,
    which allows you to define the graph by specifying the dependencies of each new
    estimator:

    >>> from skdag import DAGBuilder
    >>> dag = (
    ...     DAGBuilder()
    ...     .add_step("impute", SimpleImputer())
    ...     .add_step("vitals", "passthrough", deps={"impute": slice(0, 4)})
    ...     .add_step("blood", PCA(n_components=2, random_state=0), deps={"impute": slice(4, 10)})
    ...     .add_step("lr", LogisticRegression(random_state=0), deps=["blood", "vitals"])
    ...     .make_dag()
    ... )
    >>> print(dag.draw().strip())
    o    impute
    |\\
    o o    blood,vitals
    |/
    o    lr

    In the above examples we pass the first four columns directly to a regressor, but
    the remaining columns have dimensionality reduction applied first before being
    passed to the same regressor. Note that we can define our graph edges in two
    different ways: as a dict (if we need to select only certain columns from the source
    node) or as a simple list (if we want to simply grab all columns from all input
    nodes).

    The DAG may now be used as an estimator in its own right:

    >>> from sklearn import datasets
    >>> X, y = datasets.load_diabetes(return_X_y=True)
    >>> dag.fit_predict(X, y)
    array([...

    In an extension to the scikit-learn estimator interface, DAGs also support multiple
    inputs and multiple outputs. Let's say we want to compare two different classifiers:

    >>> from sklearn.ensemble import RandomForestClassifier
    >>> cal = DAG.from_pipeline(
    ...     [("rf", RandomForestClassifier(random_state=0))]
    ... )
    >>> dag2 = dag.join(cal, edges=[("blood", "rf"), ("vitals", "rf")])
    >>> print(dag2.draw().strip())
    o    impute
    |\\
    o o    blood,vitals
    |x|
    o o    lr,rf

    Now our DAG will return two outputs: one from each classifier. Multiple outputs are
    returned as a :class:`sklearn.utils.Bunch<Bunch>`:

    >>> y_pred = dag2.fit_predict(X, y)
    >>> y_pred.lr
    array([...
    >>> y_pred.rf
    array([...

    Similarly, multiple inputs are also acceptable and inputs can be provided by
    specifying ``X`` and ``y`` as a ``dict``-like object.
    """

    # BaseEstimator interface
    _required_parameters = ["graph"]

    @classmethod
    @deprecated(
        "DAG.from_pipeline is deprecated in 0.0.3 and will be removed in a future "
        "release. Please use DAGBuilder.from_pipeline instead."
    )
    def from_pipeline(cls, steps, **kwargs):
        from skdag.dag._builder import DAGBuilder

        return DAGBuilder().from_pipeline(steps, **kwargs).make_dag()

    def __init__(self, graph, *, memory=None, n_jobs=None, verbose=False):
        self.graph = graph
        self.memory = memory
        self.verbose = verbose
        self.n_jobs = n_jobs

    def get_params(self, deep=True):
        """
        Get parameters for this metaestimator.

        Returns the parameters given in the constructor as well as the
        estimators contained within the `steps_` of the `DAG`.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        return self._get_params("steps_", deep=deep)

    def set_params(self, **params):
        """
        Set the parameters of this metaestimator.

        Valid parameter keys can be listed with ``get_params()``. Note that
        you can directly set the parameters of the estimators contained in
        `steps_`.

        Parameters
        ----------
        **params : dict
            Parameters of this metaestimator or parameters of estimators contained
            in `steps`. Parameters of the steps may be set using its name and
            the parameter name separated by a '__'.

        Returns
        -------
        self : object
            DAG class instance.
        """
        step_names = set(self.step_names)
        for param in list(params.keys()):
            if "__" not in param and param in step_names:
                self.graph_.nodes[param]["step"].estimator = params.pop(param)

        super().set_params(**params)
        return self

    def _log_message(self, step):
        if not self.verbose:
            return None

        return f"(step {step.name}: {step.index} of {len(self.graph_)}) Processing {step.name}"

    def _iter(self, with_leaves=True, filter_passthrough=True):
        """
        Generate stage lists from self.graph_.
        When filter_passthrough is True, 'passthrough' and None transformers
        are filtered out.
        """
        for stage in nx.topological_generations(self.graph_):
            stage = [self.graph_.nodes[step]["step"] for step in stage]
            if not with_leaves:
                stage = [step for step in stage if not step.is_leaf]

            if filter_passthrough:
                stage = [
                    step
                    for step in stage
                    if step.estimator is not None and step.estimator != "passthough"
                ]

            if len(stage) == 0:
                continue

            yield stage

    def __len__(self):
        """
        Returns the size of the DAG
        """
        return len(self.graph_)

    def __getitem__(self, name):
        """
        Retrieve a named estimator.
        """
        return self.graph_.nodes[name]["step"].estimator

    def _fit(self, X, y=None, **fit_params_steps):
        # Setup the memory
        memory = check_memory(self.memory)

        fit_transform_one_cached = memory.cache(_fit_transform_one)

        root_names = set([root.name for root in self.roots_])
        Xin = self._resolve_inputs(X)
        Xs = {}
        with Parallel(n_jobs=self.n_jobs) as parallel:
            for stage in self._iter(with_leaves=False, filter_passthrough=False):
                stage_names = [step.name for step in stage]
                outputs, fitted_transformers = zip(
                    *parallel(
                        delayed(_parallel_fit)(
                            self,
                            step,
                            Xin,
                            Xs,
                            y,
                            fit_transform_one_cached,
                            memory,
                            **fit_params_steps[step.name],
                        )
                        for step in stage
                    )
                )

                for step, fitted_transformer in zip(stage, fitted_transformers):
                    # Replace the transformer of the step with the fitted
                    # transformer. This is necessary when loading the transformer
                    # from the cache.
                    step.estimator = fitted_transformer
                    step.is_fitted = True

                Xs.update(dict(zip(stage_names, outputs)))

                # If all of a dep's dependents are now complete, we can free up some
                # memory.
                root_names = root_names - set(stage_names)
                for dep in {dep for step in stage for dep in step.deps}:
                    dependents = self.graph_.successors(dep)
                    if all(d in Xs and d not in root_names for d in dependents):
                        del Xs[dep]

        # If a root node is also a leaf, it hasn't been fit yet and we need to pass on
        # its input for later.
        Xs.update({name: Xin[name] for name in root_names})
        return Xs

    def _transform(self, X, **fn_params_steps):
        # Setup the memory
        memory = check_memory(self.memory)

        transform_one_cached = memory.cache(_transform_one)

        root_names = set([root.name for root in self.roots_])
        Xin = self._resolve_inputs(X)
        Xs = {}
        with Parallel(n_jobs=self.n_jobs) as parallel:
            for stage in self._iter(with_leaves=False, filter_passthrough=False):
                stage_names = [step.name for step in stage]
                outputs = parallel(
                    delayed(_parallel_transform)(
                        self,
                        step,
                        Xin,
                        Xs,
                        transform_one_cached,
                        **fn_params_steps[step.name],
                    )
                    for step in stage
                )

                Xs.update(dict(zip(stage_names, outputs)))

                # If all of a dep's dependents are now complete, we can free up some
                # memory.
                root_names = root_names - set(stage_names)
                for dep in {dep for step in stage for dep in step.deps}:
                    dependents = self.graph_.successors(dep)
                    if all(d in Xs and d not in root_names for d in dependents):
                        del Xs[dep]

        # If a root node is also a leaf, it hasn't been fit yet and we need to pass on
        # its input for later.
        Xs.update({name: Xin[name] for name in root_names})
        return Xs

    def _resolve_inputs(self, X):
        if isinstance(X, (dict, Bunch, UserDict)) and not isinstance(X, dok_matrix):
            inputs = sorted(X.keys())
            if inputs != sorted(root.name for root in self.roots_):
                raise ValueError(
                    "Input dicts must contain one key per entry node. "
                    f"Entry nodes are {self.roots_}, got {inputs}."
                )
        else:
            if len(self.roots_) != 1:
                raise ValueError(
                    "Must provide a dictionary of inputs for a DAG with multiple entry "
                    "points."
                )
            X = {self.roots_[0].name: X}

        X = {
            step: x if issparse(x) or _is_pandas(x) else np.asarray(x)
            for step, x in X.items()
        }

        return X

    def _match_input_format(self, Xin, Xout):
        if len(self.leaves_) == 1 and (
            not isinstance(Xin, (dict, Bunch, UserDict)) or isinstance(Xin, dok_matrix)
        ):
            return Xout[self.leaves_[0].name]
        return Bunch(**Xout)

    def fit(self, X, y=None, **fit_params):
        """
        Fit the model.

        Fit all the transformers one after the other and transform the
        data. Finally, fit the transformed data using the final estimators.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            DAG.
        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the DAG.
        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        self : object
            DAG fitted steps.
        """
        self._validate_graph()
        fit_params_steps = self._check_fit_params(**fit_params)
        Xts = self._fit(X, y, **fit_params_steps)
        fitted_estimators = Parallel(n_jobs=self.n_jobs)(
            [
                delayed(_parallel_fit_leaf)(
                    self, leaf, Xts, y, **fit_params_steps[leaf.name]
                )
                for leaf in self.leaves_
            ]
        )
        for est, leaf in zip(fitted_estimators, self.leaves_):
            leaf.estimator = est
            leaf.is_fitted = True

        # If we have a single root, mirror certain attributes in the DAG.
        if len(self.roots_) == 1:
            root = self.roots_[0].estimator
            for attr in ["n_features_in_", "feature_names_in_"]:
                if hasattr(root, attr):
                    setattr(self, attr, getattr(root, attr))

        return self

    def _fit_execute(self, fn, X, y=None, **fit_params):
        self._validate_graph()
        fit_params_steps = self._check_fit_params(**fit_params)
        Xts = self._fit(X, y, **fit_params_steps)
        Xout = {}

        leaf_names = [leaf.name for leaf in self.leaves_]
        outputs, fitted_estimators = zip(
            *Parallel(n_jobs=self.n_jobs)(
                delayed(_parallel_execute)(
                    self,
                    leaf,
                    fn,
                    Xts,
                    y,
                    fit_first=True,
                    fit_params=fit_params_steps[leaf.name],
                )
                for leaf in self.leaves_
            )
        )

        Xout = dict(zip(leaf_names, outputs))
        for step, fitted_estimator in zip(self.leaves_, fitted_estimators):
            step.estimator = fitted_estimator
            step.is_fitted = True

        return self._match_input_format(X, Xout)

    def _execute(self, fn, X, y=None, **fn_params):
        Xout = {}
        fn_params_steps = self._check_fn_params(**fn_params)
        Xts = self._transform(X, **fn_params_steps)

        leaf_names = [leaf.name for leaf in self.leaves_]
        outputs, _ = zip(
            *Parallel(n_jobs=self.n_jobs)(
                delayed(_parallel_execute)(
                    self,
                    leaf,
                    fn,
                    Xts,
                    y,
                    fit_first=False,
                    fn_params=fn_params_steps[leaf.name],
                )
                for leaf in self.leaves_
            )
        )

        Xout = dict(zip(leaf_names, outputs))

        return self._match_input_format(X, Xout)

    @available_if(_leaf_estimators_have("transform"))
    def fit_transform(self, X, y=None, **fit_params):
        """
        Fit the model and transform with the final estimator.

        Fits all the transformers one after the other and transform the
        data. Then uses `fit_transform` on transformed data with the final
        estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            DAG.
        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the DAG.
        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed samples.
        """
        return self._fit_execute("transform", X, y, **fit_params)

    @available_if(_leaf_estimators_have("transform"))
    def transform(self, X):
        """
        Transform the data, and apply `transform` with the final estimator.

        Call `transform` of each transformer in the DAG. The transformed
        data are finally passed to the final estimator that calls
        `transform` method. Only valid if the final estimator
        implements `transform`.

        This also works where final estimator is `None` in which case all prior
        transformations are applied.

        Parameters
        ----------
        X : iterable
            Data to transform. Must fulfill input requirements of first step
            of the DAG.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed data.
        """
        return self._execute("transform", X)

    @available_if(_leaf_estimators_have("predict"))
    def fit_predict(self, X, y=None, **fit_params):
        """
        Transform the data, and apply `fit_predict` with the final estimator.

        Call `fit_transform` of each transformer in the DAG. The transformed data are
        finally passed to the final estimator that calls `fit_predict` method. Only
        valid if the final estimators implement `fit_predict`.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of
            the DAG.
        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps
            of the DAG.
        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        y_pred : ndarray
            Result of calling `fit_predict` on the final estimator.
        """
        return self._fit_execute("predict", X, y, **fit_params)

    @available_if(_leaf_estimators_have("predict"))
    def predict(self, X, **predict_params):
        """
        Transform the data, and apply `predict` with the final estimator.

        Call `transform` of each transformer in the DAG. The transformed
        data are finally passed to the final estimator that calls `predict`
        method. Only valid if the final estimators implement `predict`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the DAG.
        **predict_params : dict of string -> object
            Parameters to the ``predict`` called at the end of all
            transformations in the DAG. Note that while this may be
            used to return uncertainties from some models with return_std
            or return_cov, uncertainties that are generated by the
            transformations in the DAG are not propagated to the
            final estimator.

        Returns
        -------
        y_pred : ndarray
            Result of calling `predict` on the final estimator.
        """
        return self._execute("predict", X, **predict_params)

    @available_if(_leaf_estimators_have("predict_proba"))
    def predict_proba(self, X, **predict_proba_params):
        """
        Transform the data, and apply `predict_proba` with the final estimator.

        Call `transform` of each transformer in the DAG. The transformed
        data are finally passed to the final estimator that calls
        `predict_proba` method. Only valid if the final estimators implement
        `predict_proba`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the DAG.
        **predict_proba_params : dict of string -> object
            Parameters to the `predict_proba` called at the end of all
            transformations in the DAG.

        Returns
        -------
        y_proba : ndarray of shape (n_samples, n_classes)
            Result of calling `predict_proba` on the final estimator.
        """
        return self._execute("predict_proba", X, **predict_proba_params)

    @available_if(_leaf_estimators_have("decision_function"))
    def decision_function(self, X):
        """
        Transform the data, and apply `decision_function` with the final estimator.

        Call `transform` of each transformer in the DAG. The transformed
        data are finally passed to the final estimator that calls
        `decision_function` method. Only valid if the final estimators
        implement `decision_function`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the DAG.

        Returns
        -------
        y_score : ndarray of shape (n_samples, n_classes)
            Result of calling `decision_function` on the final estimator.
        """
        return self._execute("decision_function", X)

    @available_if(_leaf_estimators_have("score_samples"))
    def score_samples(self, X):
        """
        Transform the data, and apply `score_samples` with the final estimator.

        Call `transform` of each transformer in the DAG. The transformed
        data are finally passed to the final estimator that calls
        `score_samples` method. Only valid if the final estimators implement
        `score_samples`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the DAG.

        Returns
        -------
        y_score : ndarray of shape (n_samples,)
            Result of calling `score_samples` on the final estimator.
        """
        return self._execute("score_samples", X)

    @available_if(_leaf_estimators_have("score"))
    def score(self, X, y=None, **score_params):
        """
        Transform the data, and apply `score` with the final estimator.

        Call `transform` of each transformer in the DAG. The transformed
        data are finally passed to the final estimator that calls
        `score` method. Only valid if the final estimators implement `score`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the DAG.
        y : iterable, default=None
            Targets used for scoring. Must fulfill label requirements for all
            steps of the DAG.
        sample_weight : array-like, default=None
            If not None, this argument is passed as ``sample_weight`` keyword
            argument to the ``score`` method of the final estimator.

        Returns
        -------
        score : float
            Result of calling `score` on the final estimator.
        """
        return self._execute("score", X, y, **score_params)

    @available_if(_leaf_estimators_have("predict_log_proba"))
    def predict_log_proba(self, X, **predict_log_proba_params):
        """
        Transform the data, and apply `predict_log_proba` with the final estimator.

        Call `transform` of each transformer in the DAG. The transformed
        data are finally passed to the final estimator that calls
        `predict_log_proba` method. Only valid if the final estimator
        implements `predict_log_proba`.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the DAG.
        **predict_log_proba_params : dict of string -> object
            Parameters to the ``predict_log_proba`` called at the end of all
            transformations in the DAG.

        Returns
        -------
        y_log_proba : ndarray of shape (n_samples, n_classes)
            Result of calling `predict_log_proba` on the final estimator.
        """
        return self._execute("predict_log_proba", X, **predict_log_proba_params)

    def _check_fit_params(self, **fit_params):
        fit_params_steps = {
            name: {} for (name, step) in self.steps_ if step is not None
        }
        for pname, pval in fit_params.items():
            if pval is None:
                continue

            if "__" not in pname:
                raise ValueError(
                    f"DAG.fit does not accept the {pname} parameter. "
                    "You can pass parameters to specific steps of your "
                    "DAG using the stepname__parameter format, e.g. "
                    "`DAG.fit(X, y, logisticregression__sample_weight"
                    "=sample_weight)`."
                )
            step, param = pname.split("__", 1)
            fit_params_steps[step][param] = pval
        return fit_params_steps

    def _check_fn_params(self, **fn_params):
        global_params = {}
        fn_params_steps = {name: {} for (name, step) in self.steps_ if step is not None}
        for pname, pval in fn_params.items():
            if pval is None:
                continue

            if "__" not in pname:
                global_params[pname] = pval
            else:
                step, param = pname.split("__", 1)
                fn_params_steps[step][param] = pval

        for step in fn_params_steps:
            fn_params_steps[step].update(global_params)

        return fn_params_steps

    def _validate_graph(self):
        if len(self.graph_) == 0:
            raise ValueError("DAG has no nodes.")

        for i, (name, est) in enumerate(self.steps_):
            step = self.graph_.nodes[name]["step"]
            step.index = i

        # validate names
        self._validate_names([name for (name, step) in self.steps_])

        # validate transformers
        for step in self.roots_ + self.branches_:
            if step in self.leaves_:
                # This will get validated later
                continue

            est = step.estimator
            # Unlike pipelines we also allow predictors to be used as a transformer, to support
            # model stacking.
            if (
                not _is_passthrough(est)
                and not _is_transformer(est)
                and not _is_predictor(est)
            ):
                raise TypeError(
                    "All intermediate steps should be "
                    "transformers and implement fit and transform "
                    "or be the string 'passthrough' "
                    f"'{est}' (type {type(est)}) doesn't"
                )

        # Validate final estimator(s)
        for step in self.leaves_:
            est = step.estimator
            if not _is_passthrough(est) and not hasattr(est, "fit"):
                raise TypeError(
                    "Leaf nodes of a DAG should implement fit "
                    "or be the string 'passthrough'. "
                    f"'{est}' (type {type(est)}) doesn't"
                )

    @property
    def graph_(self):
        if not hasattr(self, "_graph"):
            # Read-only view of the graph. We should not modify
            # the original graph.
            self._graph = self.graph.copy(as_view=True)

        return self._graph

    @property
    def leaves_(self):
        if not hasattr(self, "_leaves"):
            self._leaves = [node for node in self.nodes_ if node.is_leaf]

        return self._leaves

    @property
    def branches_(self):
        if not hasattr(self, "_branches"):
            self._branches = [
                node for node in self.nodes_ if not node.is_leaf and not node.is_root
            ]

        return self._branches

    @property
    def roots_(self):
        if not hasattr(self, "_roots"):
            self._roots = [node for node in self.nodes_ if node.is_root]

        return self._roots

    @property
    def nodes_(self):
        if not hasattr(self, "_nodes"):
            self._nodes = []
            for name, estimator in self.steps_:
                step = self.graph_.nodes[name]["step"]
                if self.graph_.out_degree(name) == 0:
                    step.is_leaf = True
                if self.graph_.in_degree(name) == 0:
                    step.is_root = True
                self._nodes.append(step)

        return self._nodes

    @property
    def steps_(self):
        "return list of (name, estimator) tuples to conform with Pipeline interface."
        if not hasattr(self, "_steps"):
            self._steps = [
                (node, self.graph_.nodes[node]["step"].estimator)
                for node in nx.lexicographical_topological_sort(self.graph_)
            ]

        return self._steps

    def join(self, other, edges, **kwargs):
        """
        Create a new DAG by joining this DAG to another one, according to the edges
        specified.

        Parameters
        ----------

        other : :class:`skdag.dag.DAG`
            The other DAG to connect to.
        edges : (str, str) or (str, str, index-like)
            ``(u, v)`` edges that connect the two DAGs. ``u`` and ``v`` should be the
            names of steps in the first and second DAG respectively. Optionally a third
            parameter may be included to specify which columns to pass along the edge.
        **kwargs : keyword params
            Any other parameters to pass to the new DAG's constructor.

        Returns
        -------
        dag : :class:`skdag.DAG`
            A new DAG, containing a copy of each of the input DAGs, joined by the
            specified edges. Note that the original input dags are unmodified.

        Examples
        --------

        >>> from sklearn.decomposition import PCA
        >>> from sklearn.impute import SimpleImputer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from sklearn.calibration import CalibratedClassifierCV
        >>> from skdag.dag import DAGBuilder
        >>> dag1 = (
        ...     DAGBuilder()
        ...     .add_step("impute", SimpleImputer())
        ...     .add_step("vitals", "passthrough", deps={"impute": slice(0, 4)})
        ...     .add_step("blood", PCA(n_components=2, random_state=0), deps={"impute": slice(4, 10)})
        ...     .add_step("lr", LogisticRegression(random_state=0), deps=["blood", "vitals"])
        ...     .make_dag()
        ... )
        >>> print(dag1.draw().strip())
        o    impute
        |\\
        o o    blood,vitals
        |/
        o    lr
        >>> dag2 = (
        ...     DAGBuilder()
        ...     .add_step(
        ...         "calib",
        ...         CalibratedClassifierCV(LogisticRegression(random_state=0), cv=5),
        ...     )
        ...     .make_dag()
        ... )
        >>> print(dag2.draw().strip())
        o    calib
        >>> dag3 = dag1.join(dag2, edges=[("blood", "calib"), ("vitals", "calib")])
        >>> print(dag3.draw().strip())
        o    impute
        |\\
        o o    blood,vitals
        |x|
        o o    calib,lr
        """
        if set(self.step_names) & set(other.step_names):
            raise ValueError("DAGs with overlapping step names cannot be combined.")

        newgraph = deepcopy(self.graph_).copy()
        for edge in edges:
            if len(edge) == 2:
                u, v, idx = *edge, None
            else:
                u, v, idx = edge

            if u not in self.graph_:
                raise KeyError(u)
            if v not in other.graph_:
                raise KeyError(v)

            # source node can no longer be a leaf
            ustep = newgraph.nodes[u]["step"]
            if ustep.is_leaf:
                ustep.is_leaf = False

            vnode = other.graph_.nodes[v]
            old_step = vnode["step"]
            vstep = DAGStep(
                name=old_step.name,
                estimator=old_step.estimator,
                deps=old_step.deps,
                dataframe_columns=old_step.dataframe_columns,
                axis=old_step.axis,
            )

            if u not in vstep.deps:
                vstep.deps[u] = idx

            vnode["step"] = vstep

            newgraph.add_node(v, **vnode)
            newgraph.add_edge(u, v)

        return DAG(newgraph, **kwargs)

    def draw(
        self, filename=None, style=None, detailed=False, format=None, layout="dot"
    ):
        """
        Render a graphical view of the DAG.

        By default the rendered file will be returned as a string. However if an output
        file is provided then the output will be saved to file.

        Parameters
        ----------

        filename : str
            The file to write the image to. If None, the rendered image will be sent to
            stdout.
        style : str, optional, choice of ['light', 'dark']
            Draw the image in light or dark mode.
        detailed : bool, default = False
            If True, show extra details in the node labels such as the estimator
            signature.
        format : str, choice of ['svg', 'png', 'jpg', 'txt']
            The rendering format to use. MAy be omitted if the format can be inferred
            from the filename.
        layout : str, default = 'dot'
            The program to use for generating a graph layout.

        See Also
        --------

        :meth:`skdag.dag.DAG.show`, for use in interactive notebooks.

        Returns
        -------

        output : str, bytes or None
            If a filename is provided the output is written to file and `None` is
            returned. Otherwise, the output is returned as a string (for textual formats
            like ascii or svg) or bytes.
        """
        if filename is None and format is None:
            try:
                from IPython import get_ipython

                rich = type(get_ipython()).__name__ == "ZMQInteractiveShell"
            except (ModuleNotFoundError, NameError):
                rich = False

            format = "svg" if rich else "txt"

        if format is None:
            format = filename.split(".")[-1]

        if format not in ["svg", "png", "jpg", "txt"]:
            raise ValueError(f"Unsupported file format '{format}'")

        render = DAGRenderer(self.graph_, detailed=detailed, style=style).draw(
            format=format, layout=layout
        )
        if filename is None:
            return render
        else:
            mode = "wb" if isinstance(render, bytes) else "w"
            with open(filename, mode) as fp:
                fp.write(render)

    def show(self, style=None, detailed=False, format=None, layout="dot"):
        """
        Display a graphical representation of the DAG in an interactive notebook
        environment.

        DAGs will be shown when displayed in a notebook, but calling this method
        directly allows more options to be passed to customise the appearance more.

        Arguments are as for :meth`.draw`.

        Returns
        -------

        ``None``

        See Also
        --------

        :meth:`skdag.DAG.draw`
        """
        if format is None:
            format = "svg" if _in_notebook() else "txt"

        data = self.draw(style=style, detailed=detailed, format=format, layout=layout)
        if format == "svg":
            from IPython.display import SVG, display

            display(SVG(data))
        elif format == "txt":
            print(data)
        elif format in ("jpg", "png"):
            from IPython.display import Image, display

            display(Image(data))
        else:
            raise ValueError(f"'{format}' format not supported.")

    def _repr_svg_(self):
        return self.draw(format="svg")

    def _repr_png_(self):
        return self.draw(format="png")

    def _repr_jpeg_(self):
        return self.draw(format="jpg")

    def _repr_html_(self):
        return self.draw(format="svg")

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(repr(self))
        else:
            p.text(str(self))

    def _repr_mimebundle_(self, **kwargs):
        # Don't render yet...
        renderers = {
            "image/svg+xml": self._repr_svg_,
            "image/png": self._repr_png_,
            "image/jpeg": self._repr_jpeg_,
            "text/plain": self.__str__,
            "text/html": self._repr_html_,
        }

        include = kwargs.get("include")
        if include:
            renderers = {k: v for k, v in renderers.items() if k in include}

        exclude = kwargs.get("exclude")
        if exclude:
            renderers = {k: v for k, v in renderers.items() if k not in exclude}

        # Now render any remaining options.
        return {k: v() for k, v in renderers.items()}

    @property
    def named_steps(self):
        """
        Access the steps by name.

        Read-only attribute to access any step by given name.
        Keys are steps names and values are the steps objects.
        """
        # Use Bunch object to improve autocomplete
        return Bunch(**dict(self.steps_))

    @property
    def step_names(self):
        return list(self.graph_.nodes)

    @property
    def edges(self):
        return self.graph_.edges

    def _get_leaf_attr(self, attr):
        if len(self.leaves_) == 1:
            return getattr(self.leaves_[0].estimator, attr)
        else:
            return Bunch(
                **{leaf.name: getattr(leaf.estimator, attr) for leaf in self.leaves_}
            )

    @property
    def _estimator_type(self):
        return self._get_leaf_attr("_estimator_type")

    @property
    def classes_(self):
        """The classes labels. Only exist if the leaf steps are classifiers."""
        return self._get_leaf_attr("classes_")

    def __sklearn_is_fitted__(self):
        """Indicate whether DAG has been fit."""
        try:
            # check if the last steps of the DAG are fitted
            # we only check the last steps since if the last steps are fit, it
            # means the previous steps should also be fit. This is faster than
            # checking if every step of the DAG is fit.
            for leaf in self._leaves:
                check_is_fitted(leaf.estimator)
            return True
        except NotFittedError:
            return False

    def _more_tags(self):
        tags = {}

        # We assume the DAG can handle NaN if *all* the steps can.
        tags["allow_nan"] = all(
            _safe_tags(node.estimator, "allow_nan") for node in self.nodes_
        )

        # Check if all *root* nodes expect pairwise input.
        tags["pairwise"] = all(
            _safe_tags(root.estimator, "pairwise") for root in self.roots_
        )

        # CHeck if all *leaf* notes support multioutput
        tags["multioutput"] = all(
            _safe_tags(leaf.estimator, "multioutput") for leaf in self.leaves_
        )

        return tags



import numpy as np
from scipy import sparse

try:
    import pandas as pd
except ImportError:
    pd = None


def _is_passthrough(estimator):
    return estimator is None or estimator == "passthrough"


def _is_transformer(estimator):
    return (
        hasattr(estimator, "fit") or hasattr(estimator, "fit_transform")
    ) and hasattr(estimator, "transform")


def _is_predictor(estimator):
    return (hasattr(estimator, "fit") or hasattr(estimator, "fit_predict")) and hasattr(
        estimator, "predict"
    )


def _in_notebook():
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def _stack(Xs, axis=0):
    """
    Where an estimator has multiple upstream dependencies, this method defines the
    strategy for merging the upstream outputs into a single input. ``axis=0`` is
    equivalent to a vstack (combination of samples) and ``axis=1`` is equivalent to a
    hstack (combination of features). Higher axes are only supported for non-sparse
    data sources.
    """
    if any(sparse.issparse(x) for x in Xs):
        if -2 <= axis < 2:
            axis = axis % 2
        else:
            raise NotImplementedError(
                f"Stacking is not supported for sparse inputs on axis > 1."
            )

        if axis == 0:
            Xs = sparse.vstack(Xs).tocsr()
        elif axis == 1:
            Xs = sparse.hstack(Xs).tocsr()
    elif pd and all(_is_pandas(x) for x in Xs):
        Xs = pd.concat(Xs, axis=axis)
    else:
        if axis == 1:
            Xs = np.hstack(Xs)
        else:
            Xs = np.stack(Xs, axis=axis)

    return Xs


def _is_pandas(X):
    "Check if X is a DataFrame or Series"
    return hasattr(X, "iloc")


def _get_feature_names(estimator):
    try:
        feature_names = estimator.get_feature_names_out()
    except AttributeError:
        try:
            feature_names = estimator.get_feature_names()
        except AttributeError:
            feature_names = None

    return feature_names


def _format_output(X, input, node):
    if node.dataframe_columns is None or pd is None or _is_pandas(X):
        return X

    outshape = np.asarray(X).shape
    outdim = len(outshape)
    if outdim > 2 or outdim < 1:
        return X

    if node.dataframe_columns == "infer":
        if outdim == 1:
            columns = [node.name]
        else:
            feature_names = _get_feature_names(node.estimator)
            if feature_names is None:
                feature_names = range(outshape[1])

            columns = [f"{node.name}__{f}" for f in feature_names]
    else:
        columns = node.dataframe_columns

    if hasattr(input, "index"):
        index = input.index
    else:
        index = None

    if outdim == 2:
        df = pd.DataFrame(X, columns=columns, index=index)
    else:
        df = pd.Series(X, name=columns[0], index=index)

    return df



import pytest
from skdag import DAG, DAGBuilder
from skdag.dag.tests.utils import Mult, Transf
from skdag.exceptions import DAGError
from sklearn.pipeline import Pipeline


def test_builder_basics():
    builder = DAGBuilder()
    builder.add_step("tr", Transf())
    builder.add_step("ident", "passthrough", deps=["tr"])
    builder.add_step("est", Mult(), deps=["ident"])

    dag = builder.make_dag()
    assert isinstance(dag, DAG)
    assert dag._repr_html_() == builder._repr_html_()

    with pytest.raises(ValueError):
        builder.add_step("est2", Mult(), deps=["missing"])

    with pytest.raises(ValueError):
        builder.add_step("est2", Mult(), deps=[1])

    with pytest.raises(KeyError):
        # Step names *must* be strings.
        builder.add_step(2, Mult())

    with pytest.raises(KeyError):
        # Step names *must* be unique.
        builder.add_step("est", Mult())

    with pytest.raises(DAGError):
        # Cycles not allowed.
        builder.graph.add_edge("est", "ident")
        builder.make_dag()


def test_pipeline():
    steps = [("tr", Transf()), ("ident", "passthrough"), ("est", Mult())]

    builder = DAGBuilder()
    prev = None
    for name, est in steps:
        builder.add_step(name, est, deps=[prev] if prev else None)
        prev = name

    dag = builder.make_dag()

    assert dag.steps_ == steps



"""
Test the DAG module.
"""
import re

import numpy as np
import pandas as pd
import pytest
from skdag import DAGBuilder
from skdag.dag.tests.utils import FitParamT, Mult, NoFit, NoTrans, Transf
from sklearn import datasets
from sklearn.base import clone
from sklearn.compose import make_column_selector
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.feature_selection import SelectKBest, f_classif
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVC
from sklearn.utils._testing import assert_array_almost_equal
from sklearn.utils.estimator_checks import parametrize_with_checks

iris = datasets.load_iris()
cancer = datasets.load_breast_cancer()

JUNK_FOOD_DOCS = (
    "the pizza pizza beer copyright",
    "the pizza burger beer copyright",
    "the the pizza beer beer copyright",
    "the burger beer beer copyright",
    "the coke burger coke copyright",
    "the coke burger burger",
)


def test_dag_invalid_parameters():
    # Test the various init parameters of the dag in fit
    # method
    with pytest.raises(KeyError):
        dag = DAGBuilder().from_pipeline([(1, 1)]).make_dag()

    # Check that we can't fit DAGs with objects without fit
    # method
    msg = (
        "Leaf nodes of a DAG should implement fit or be the string 'passthrough'"
        ".*NoFit.*"
    )
    dag = DAGBuilder().from_pipeline([("clf", NoFit())]).make_dag()
    with pytest.raises(TypeError, match=msg):
        dag.fit([[1]], [1])

    # Smoke test with only an estimator
    clf = NoTrans()
    dag = DAGBuilder().from_pipeline([("svc", clf)]).make_dag()
    assert dag.get_params(deep=True) == dict(
        svc__a=None, svc__b=None, svc=clf, **dag.get_params(deep=False)
    )

    # Check that params are set
    dag.set_params(svc__a=0.1)
    assert clf.a == 0.1
    assert clf.b is None
    # Smoke test the repr:
    repr(dag)

    # Test with two objects
    clf = SVC()
    filter1 = SelectKBest(f_classif)
    dag = DAGBuilder().from_pipeline([("anova", filter1), ("svc", clf)]).make_dag()

    # Check that estimators are not cloned on pipeline construction
    assert dag.named_steps["anova"] is filter1
    assert dag.named_steps["svc"] is clf

    # Check that we can't fit with non-transformers on the way
    # Note that NoTrans implements fit, but not transform
    msg = "All intermediate steps should be transformers.*\\bNoTrans\\b.*"
    dag2 = DAGBuilder().from_pipeline([("t", NoTrans()), ("svc", clf)]).make_dag()
    with pytest.raises(TypeError, match=msg):
        dag2.fit([[1]], [1])

    # Check that params are set
    dag.set_params(svc__C=0.1)
    assert clf.C == 0.1
    # Smoke test the repr:
    repr(dag)

    # Check that params are not set when naming them wrong
    msg = re.escape(
        "Invalid parameter 'C' for estimator SelectKBest(). Valid parameters are: ['k',"
        " 'score_func']."
    )
    with pytest.raises(ValueError, match=msg):
        dag.set_params(anova__C=0.1)

    # Test clone
    dag2 = clone(dag)
    assert not dag.named_steps["svc"] is dag2.named_steps["svc"]

    # Check that apart from estimators, the parameters are the same
    params = dag.get_params(deep=True)
    params2 = dag2.get_params(deep=True)

    for x in dag.get_params(deep=False):
        params.pop(x)

    for x in dag.get_params(deep=False):
        params2.pop(x)

    # Remove estimators that where copied
    params.pop("svc")
    params.pop("anova")
    params2.pop("svc")
    params2.pop("anova")
    assert params == params2


def test_dag_simple():
    # Build a simple DAG of one node
    rng = np.random.default_rng(1)
    X = rng.random(size=(20, 5))

    dag = DAGBuilder().add_step("clf", FitParamT()).make_dag()
    dag.fit(X)
    dag.predict(X)


def test_dag_pipeline_init():
    # Build a dag from a pipeline
    X = np.array([[1, 2]])
    steps = (("transf", Transf()), ("clf", FitParamT()))
    pipe = Pipeline(steps, verbose=False)
    for inp in [pipe, steps]:
        dag = DAGBuilder().from_pipeline(inp).make_dag()
        dag.fit(X, y=None)
        dag.score(X)

        dag.set_params(transf="passthrough")
        dag.fit(X, y=None)
        dag.score(X)


def test_dag_methods_anova():
    # Test the various methods of the dag (anova).
    X = iris.data
    y = iris.target
    # Test with Anova + LogisticRegression
    clf = LogisticRegression()
    filter1 = SelectKBest(f_classif, k=2)
    dag1 = (
        DAGBuilder().from_pipeline([("anova", filter1), ("logistic", clf)]).make_dag()
    )
    dag2 = (
        DAGBuilder()
        .add_step("anova", filter1)
        .add_step("logistic", clf, deps=["anova"])
        .make_dag()
    )
    dag1.fit(X, y)
    dag2.fit(X, y)
    assert_array_almost_equal(dag1.predict(X), dag2.predict(X))
    assert_array_almost_equal(dag1.predict_proba(X), dag2.predict_proba(X))
    assert_array_almost_equal(dag1.predict_log_proba(X), dag2.predict_log_proba(X))
    assert_array_almost_equal(dag1.score(X, y), dag2.score(X, y))


def test_dag_fit_params():
    # Test that the pipeline can take fit parameters
    dag = (
        DAGBuilder()
        .from_pipeline([("transf", Transf()), ("clf", FitParamT())])
        .make_dag()
    )
    dag.fit(X=None, y=None, clf__should_succeed=True)
    # classifier should return True
    assert dag.predict(None)
    # and transformer params should not be changed
    assert dag.named_steps["transf"].a is None
    assert dag.named_steps["transf"].b is None
    # invalid parameters should raise an error message

    msg = re.escape("fit() got an unexpected keyword argument 'bad'")
    with pytest.raises(TypeError, match=msg):
        dag.fit(None, None, clf__bad=True)


def test_dag_sample_weight_supported():
    # DAG should pass sample_weight
    X = np.array([[1, 2]])
    dag = (
        DAGBuilder()
        .from_pipeline([("transf", Transf()), ("clf", FitParamT())])
        .make_dag()
    )
    dag.fit(X, y=None)
    assert dag.score(X) == 3
    assert dag.score(X, y=None) == 3
    assert dag.score(X, y=None, sample_weight=None) == 3
    assert dag.score(X, sample_weight=np.array([2, 3])) == 8


def test_dag_sample_weight_unsupported():
    # When sample_weight is None it shouldn't be passed
    X = np.array([[1, 2]])
    dag = DAGBuilder().from_pipeline([("transf", Transf()), ("clf", Mult())]).make_dag()
    dag.fit(X, y=None)
    assert dag.score(X) == 3
    assert dag.score(X, sample_weight=None) == 3

    msg = re.escape("score() got an unexpected keyword argument 'sample_weight'")
    with pytest.raises(TypeError, match=msg):
        dag.score(X, sample_weight=np.array([2, 3]))


def test_dag_raise_set_params_error():
    # Test dag raises set params error message for nested models.
    dag = DAGBuilder().from_pipeline([("cls", LinearRegression())]).make_dag()

    # expected error message
    error_msg = (
        r"Invalid parameter 'fake' for estimator DAG\(graph=<networkx\..*DiGraph[^>]*>"
        r"\)\. Valid parameters are: \['graph', 'memory', 'n_jobs', 'verbose'\]."
    )
    with pytest.raises(ValueError, match=error_msg):
        dag.set_params(fake="nope")

    # invalid outer parameter name for compound parameter: the expected error message
    # is the same as above.
    with pytest.raises(ValueError, match=error_msg):
        dag.set_params(fake__estimator="nope")

    # expected error message for invalid inner parameter
    error_msg = (
        r"Invalid parameter 'invalid_param' for estimator LinearRegression\(\)\. Valid"
        r" parameters are: \['copy_X', 'fit_intercept', 'n_jobs', 'normalize',"
        r" 'positive'\]."
    )
    with pytest.raises(ValueError, match=error_msg):
        dag.set_params(cls__invalid_param="nope")


@pytest.mark.parametrize("idx", [1, [1]])
def test_dag_stacking_pca_svm_rf(idx):
    # Test the various methods of the pipeline (pca + svm).
    X = cancer.data
    y = cancer.target
    # Build a simple model stack with some preprocessing.
    pca = PCA(svd_solver="full", n_components="mle", whiten=True)
    svc = SVC(probability=True, random_state=0)
    rf = RandomForestClassifier(random_state=0)
    log = LogisticRegression()

    dag = (
        DAGBuilder()
        .add_step("pca", pca)
        .add_step("svc", svc, deps=["pca"])
        .add_step("rf", rf, deps=["pca"])
        .add_step("log", log, deps={"svc": idx, "rf": idx})
        .make_dag()
    )
    dag.fit(X, y)

    prob_shape = len(cancer.target), len(cancer.target_names)
    tgt_shape = cancer.target.shape

    assert dag.predict_proba(X).shape == prob_shape
    assert dag.predict(X).shape == tgt_shape
    assert dag.predict_log_proba(X).shape == prob_shape
    assert isinstance(dag.score(X, y), (float, np.floating))

    root = dag["log"]
    for attr in ["n_features_in_", "feature_names_in_"]:
        if hasattr(root, attr):
            assert hasattr(dag, attr)


def test_dag_draw():
    txt = DAGBuilder().make_dag().draw(format="txt")
    assert "[empty]" in txt

    svg = DAGBuilder().make_dag().draw(format="svg")
    assert "<title>[empty]</title>" in svg

    # Build a simple model stack with some preprocessing.
    pca = PCA(svd_solver="full", n_components="mle", whiten=True)
    svc = SVC(probability=True, random_state=0)
    rf = RandomForestClassifier(random_state=0)
    log = LogisticRegression()

    dag = (
        DAGBuilder()
        .add_step("pca", pca)
        .add_step("svc1", svc, deps={"pca": slice(4)})
        .add_step("svc2", svc, deps={"pca": [0, 1, 2]})
        .add_step(
            "svc3", svc, deps={"pca": lambda X: [c for c in X if c.startswith("foo")]}
        )
        .add_step("rf1", rf, deps={"pca": [0, 1, 2, 3, 4, 5, 6, 7, 8]})
        .add_step("rf2", rf, deps={"pca": make_column_selector(pattern="^pca.*")})
        .add_step("log", log, deps=["svc1", "svc2", "rf1", "rf2"])
        .make_dag()
    )

    for repr_method in [fn for fn in dir(dag) if fn.startswith("_repr_")]:
        if repr_method == "_repr_pretty_":
            try:
                from IPython.lib.pretty import PrettyPrinter
            except ImportError:  # pragma: no cover
                continue
            from io import StringIO

            sio = StringIO()
            getattr(dag, repr_method)(PrettyPrinter(sio), False)
            sio.seek(0)
            out = sio.read()
        else:
            out = getattr(dag, repr_method)()

        if repr_method == "_repr_mimebundle_":
            for mimetype, data in out.items():
                if mimetype in ("image/png", "image/jpeg"):
                    expected = bytes
                else:
                    expected = str
                assert isinstance(
                    data, expected
                ), f"{repr_method} {mimetype} returns unexpected type {data}"
        elif repr_method in ("_repr_jpeg_", "_repr_png_"):
            assert isinstance(
                out, bytes
            ), f"{repr_method} returns unexpected type {out}"
        else:
            assert isinstance(out, str), f"{repr_method} returns unexpected type {out}"

    txt = dag.draw(format="txt")
    for step in dag.step_names:
        assert step in txt

    svg = dag.draw(format="svg")
    for step in dag.step_names:
        assert f"<title>{step}</title>" in svg

    svg = dag.draw(format="svg", style="dark")
    for step in dag.step_names:
        assert f"<title>{step}</title>" in svg

    with pytest.raises(ValueError):
        badstyle = "foo"
        dag.draw(format="svg", style=badstyle)

    svg = dag.draw(format="svg", detailed=True)
    for step, est in dag.steps_:
        assert f"<title>{step}</title>" in svg
        assert f"{type(est).__name__}" in svg


def _dag_from_steplist(steps, **builder_opts):
    builder = DAGBuilder(**builder_opts)
    for step in steps:
        builder.add_step(**step)
    return builder.make_dag()


@pytest.mark.parametrize(
    "steps",
    [
        [
            {
                "name": "pca",
                "est": PCA(n_components=1),
            },
            {
                "name": "svc",
                "est": SVC(probability=True, random_state=0),
                "deps": ["pca"],
            },
            {
                "name": "rf",
                "est": RandomForestClassifier(random_state=0),
                "deps": ["pca"],
            },
            {
                "name": "log",
                "est": LogisticRegression(),
                "deps": ["svc", "rf"],
            },
        ],
    ],
)
@pytest.mark.parametrize(
    "X,y",
    [datasets.make_blobs(n_samples=200, n_features=10, centers=3, random_state=0)],
)
def test_pandas(X, y, steps):
    dag_np = _dag_from_steplist(steps, infer_dataframe=False)
    dag_pd = _dag_from_steplist(steps, infer_dataframe=True)

    dag_np.fit(X, y)
    dag_pd.fit(X, y)

    y_pred_np = dag_np.predict_proba(X)
    y_pred_pd = dag_pd.predict_proba(X)
    assert isinstance(y_pred_np, np.ndarray)
    assert isinstance(y_pred_pd, pd.DataFrame)
    assert np.allclose(y_pred_np, y_pred_pd)


def test_pandas_indexing():
    X, y = datasets.load_diabetes(return_X_y=True, as_frame=True)

    passcols = ["age", "sex", "bmi", "bp"]
    preprocessing = (
        DAGBuilder(infer_dataframe=True)
        .add_step("imp", SimpleImputer())
        .add_step("vitals", "passthrough", deps={"imp": passcols})
        .add_step(
            "blood",
            PCA(n_components=2, random_state=0),
            deps={"imp": make_column_selector("s[0-9]+")},
        )
        .add_step("out", "passthrough", deps=["vitals", "blood"])
        .make_dag()
    )

    X_tr = preprocessing.fit_transform(X, y)
    assert isinstance(X_tr, pd.DataFrame)
    assert (X_tr.index == X.index).all()
    assert X_tr.columns.tolist() == [f"imp__{col}" for col in passcols] + [
        "blood__pca0",
        "blood__pca1",
    ]

    predictor = (
        DAGBuilder(infer_dataframe=True)
        .add_step("rf", RandomForestRegressor(random_state=0))
        .make_dag()
    )

    dag = preprocessing.join(
        predictor,
        edges=[("out", "rf")],
    )

    y_pred = dag.fit_predict(X, y)

    assert isinstance(y_pred, pd.Series)
    assert (y_pred.index == y.index).all()
    assert y_pred.name == dag.leaves_[0].name


@parametrize_with_checks(
    [
        DAGBuilder().from_pipeline([("ss", StandardScaler())]).make_dag(),
        DAGBuilder().from_pipeline([("lr", LinearRegression())]).make_dag(),
        (
            DAGBuilder()
            .add_step("pca", PCA(n_components=1))
            .add_step("svc", SVC(probability=True, random_state=0), deps=["pca"])
            .add_step("rf", RandomForestClassifier(random_state=0), deps=["pca"])
            .add_step("log", LogisticRegression(), deps=["svc", "rf"])
            .make_dag()
        ),
    ]
)
def test_dag_check_estimator(estimator, check):
    # Since some parameters are estimators, we expect them to be modified during fit(),
    # which is why we skip these checks (in line with checks on other metaestimators
    # like Pipeline)
    if check.func.__name__ in [
        "check_estimators_overwrite_params",
        "check_dont_overwrite_parameters",
    ]:
        # we don't clone in pipeline or feature union
        return
    check(estimator)



from time import time

import numpy as np
from sklearn.base import BaseEstimator


class NoFit:  # pragma: no cover
    """Small class to test parameter dispatching."""

    def __init__(self, a=None, b=None):
        self.a = a
        self.b = b


class NoTrans(NoFit):  # pragma: no cover
    def fit(self, X, y):
        return self

    def get_params(self, deep=False):
        return {"a": self.a, "b": self.b}

    def set_params(self, **params):
        self.a = params["a"]
        return self


class NoInvTransf(NoTrans):  # pragma: no cover
    def transform(self, X):
        return X


class Transf(NoInvTransf):  # pragma: no cover
    def inverse_transform(self, X):
        return X


class TransfFitParams(Transf):  # pragma: no cover
    def fit(self, X, y, **fit_params):
        self.fit_params = fit_params
        return self


class Mult(BaseEstimator):  # pragma: no cover
    def __init__(self, mult=1):
        self.mult = mult

    def fit(self, X, y):
        return self

    def transform(self, X):
        return np.asarray(X) * self.mult

    def inverse_transform(self, X):
        return np.asarray(X) / self.mult

    def predict(self, X):
        return (np.asarray(X) * self.mult).sum(axis=1)

    predict_proba = predict_log_proba = decision_function = predict

    def score(self, X, y=None):
        return np.sum(X)


class FitParamT(BaseEstimator):  # pragma: no cover
    """Mock classifier"""

    def __init__(self):
        self.successful = False

    def fit(self, X, y, should_succeed=False):
        self.successful = should_succeed
        return self

    def predict(self, X):
        return self.successful

    def fit_predict(self, X, y, should_succeed=False):
        self.fit(X, y, should_succeed=should_succeed)
        return self.predict(X)

    def score(self, X, y=None, sample_weight=None):
        if sample_weight is not None:
            X = X * sample_weight
        return np.sum(X)


class DummyTransf(Transf):  # pragma: no cover
    """Transformer which store the column means"""

    def fit(self, X, y):
        self.means_ = np.mean(X, axis=0)
        # store timestamp to figure out whether the result of 'fit' has been
        # cached or not
        self.timestamp_ = time()
        return self


class DummyEstimatorParams(BaseEstimator):  # pragma: no cover
    """Mock classifier that takes params on predict"""

    def fit(self, X, y):
        return self

    def predict(self, X, got_attribute=False):
        self.got_attribute = got_attribute
        return self

    def predict_proba(self, X, got_attribute=False):
        self.got_attribute = got_attribute
        return self

    def predict_log_proba(self, X, got_attribute=False):
        self.got_attribute = got_attribute
        return self
